package swing;

/*
Copyright (C) 2015 Takuya KOUMURA
http://cycentum.miraiserver.com/intervalplayer/

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Attribution notices are included in the NOTICE file.
 */

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.nio.file.InvalidPathException;
import java.nio.file.Path;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;

@SuppressWarnings("serial")
public class BrowseButton extends JButton
{
	private Path directory, selected;
	private JLabel pathLabel;
	private boolean write;
	private String filterDescription;
	private String[] filterSuffix;
	private ChangeListener changeListener;
	private AudioFormat format;
	
	public BrowseButton(Path directory, JLabel pathLabel, boolean write, String filterDescription, String[] filterSuffix, ChangeListener changeListener) {
		super("Browse");
		this.directory = directory;
		this.pathLabel=pathLabel;
		this.write=write;
		this.filterDescription=filterDescription;
		this.filterSuffix=filterSuffix;
		this.changeListener=changeListener;
		addActionListener(new List(this));
	}

	public Path getSelected() {
		return selected;
	}

	public void setSelected(Path selected)
	{
		this.selected = selected;
		try
		{
			format=AudioSystem.getAudioFileFormat(selected.toFile()).getFormat();
		}
		catch(UnsupportedAudioFileException | IOException e){e.printStackTrace();}
	}

	public Path getDirectory() {
		return directory;
	}

	public void setDirectory(Path directory) {
		this.directory = directory;
	}
	
	public AudioFormat getFormat() {
		return format;
	}

	public void setFormat(AudioFormat format) {
		this.format = format;
	}

	private static class List implements ActionListener
	{
		private BrowseButton button;
		
		private List(BrowseButton button) {
			this.button = button;
		}
		
		@Override
		public void actionPerformed(ActionEvent event)
		{
			Chooser chooser=new Chooser(button.write);
			if(button.directory!=null) chooser.setCurrentDirectory(button.directory.toFile());
			if(button.filterDescription!=null) chooser.setFileFilter(new FileNameExtensionFilter(button.filterDescription, button.filterSuffix));
			int answer;
			if(button.write) answer=chooser.showSaveDialog(button);
			else answer=chooser.showOpenDialog(button);
			if(answer==JFileChooser.APPROVE_OPTION)
			{
				File file=chooser.getSelectedFile();
				Path path=chooser.getSelectedFile().toPath().toAbsolutePath();

				if(button.write)
				{
					if(!chooser.quoted)
					{
						boolean hasSuffix=false;
						for(String s: button.filterSuffix) if(path.toString().endsWith(s))
						{
							hasSuffix=true;
							break;
						}
						if(!hasSuffix) path=path.resolveSibling(path.getFileName()+"."+button.filterSuffix[0]);
						file=path.toFile();
					}
					if(file.exists())
					{
						int op=JOptionPane.showConfirmDialog(button, path.toString()+" exists. Overwrite?", "Warning", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
						if(op!=JOptionPane.OK_OPTION) return;
					}
				}
				else
				{
					AudioFormat format=null;
					try
					{
						format=AudioSystem.getAudioFileFormat(file).getFormat();
					}
					catch(UnsupportedAudioFileException e)
					{
						JOptionPane.showMessageDialog(null, "Unsupported audio file.", "Error", JOptionPane.ERROR_MESSAGE);
						return;
					}
					catch(IOException e)
					{
						JOptionPane.showMessageDialog(null, "Could not read this file.", "Error", JOptionPane.ERROR_MESSAGE);
						return;
					}
					button.format=format;
				}
				
				button.directory=path.getParent();
				button.selected=path;
				button.pathLabel.setText(path.toAbsolutePath().toString());
				button.pathLabel.setToolTipText(path.toAbsolutePath().toString());
				if(button.changeListener!=null) button.changeListener.changed(button);
			}
		}
	}
	
	private static class Chooser extends JFileChooser
	{
		private boolean quoted;
		private boolean write;
		
		private Chooser(boolean write)
		{
			super();
			this.write= write;
		}
		
		@Override
        public void approveSelection(){
            File file = getSelectedFile();
            if(write)
            {
	            String filename=file.getName();
	            if(filename.startsWith("\"")&&filename.endsWith("\""))
	            {
	            	quoted=true;
	            	filename=filename.substring(1, filename.length()-1);
	            	file=new File(file.getParentFile(), filename);
	            	setSelectedFile(file);
	            }
	            Path path=null;
	            try
	            {
	            	path=file.toPath();
	            }
	            catch(InvalidPathException e)
	            {
	            	JOptionPane.showMessageDialog(null, "File name "+file.getName()+" is invalid.", "Error", JOptionPane.ERROR_MESSAGE);
	            	return;
	            }
	            if(path==null||path.getParent()==null)
	            {
	            	JOptionPane.showMessageDialog(null, "Path "+path+" is invalid.", "Error", JOptionPane.ERROR_MESSAGE);
	            	return;
	            }
            }
            else
            {
            	if(!file.exists())
            	{
            		JOptionPane.showMessageDialog(null, file+" does not exist.", "Error", JOptionPane.ERROR_MESSAGE);
	            	return;
            	}
            }
            super.approveSelection();
        }
	}
	
	public static interface ChangeListener
	{
		void changed(BrowseButton button);
	}
}
