package swing;

/*
Copyright (C) 2015 Takuya KOUMURA
http://cycentum.miraiserver.com/intervalplayer/

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Attribution notices are included in the NOTICE file.
 */

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import main.Config;
import net.miginfocom.swing.MigLayout;
import ordering.Ordering;
import standardoutput.StandardOutputTextPane;
import utils.Console;
import utils.Utils;

@SuppressWarnings("serial")
public class MainPanel extends JPanel
{
	private MainFrame mainFrame;
	private JComboBox<String> mixerCombo;
	private StandardOutputTextPane outputText;
	private JTextField silentText, numBlockText;
	private BrowseButton logButton;
	private LinkedList<SoundFileLine> soundFileLine;
	private JPanel soundFilePanel;
	private JButton startButton;
	private boolean playing;
	private static final Pattern INTEGER_PATTERN=Pattern.compile("^\\d+$");

	public MainPanel(MainFrame mainFrame)
	{
		this.mainFrame = mainFrame;
		setLayout(new MigLayout());
		{
			JPanel leftPanel=new JPanel();
			add(leftPanel);
			leftPanel.setLayout(new MigLayout());
			{
				JPanel panel=new JPanel();
				panel.setLayout(new MigLayout());
				leftPanel.add(panel, "wrap");
				
				panel.add(new JLabel("Silent interval:"));
				silentText=new JTextField(8);
				panel.add(silentText);
				panel.add(new JLabel("ms"), "wrap");

				panel.add(new JLabel("Repeat each file"));
				numBlockText=new JTextField(8);
				panel.add(numBlockText);
				panel.add(new JLabel("times"), "wrap");
				
				panel.add(new JLabel("Log file:"));
				
				JLabel pathLabel=new JLabel();
				panel.add(pathLabel);
				pathLabel.setMaximumSize(Dimensions.PATH_LABEL_MAX);
				logButton=new BrowseButton(null, pathLabel, true, "*.txt", new String[]{"txt"}, null);
				panel.add(logButton);
			}
			{
				soundFilePanel=new JPanel();
				soundFilePanel.setLayout(new MigLayout());
				soundFilePanel.add(new JLabel("Sound files:"), "wrap");
				
				
				soundFileLine=new LinkedList<>();
				SoundFileLine line=new SoundFileLine(this, null, soundFilePanel);
				soundFileLine.add(line);
				
				JScrollPane scroll=new JScrollPane(soundFilePanel);
				leftPanel.add(scroll, "wrap");
				scroll.setBorder(null);
				scroll.setPreferredSize(Dimensions.PATH_MAX);
			}
			{
				JPanel panel=new JPanel();
				panel.setLayout(new MigLayout());
				leftPanel.add(panel, "wrap");
				
				panel.add(new JLabel("Audio devices:"));
				
				JButton reloadMixerButton=new JButton("Reload");
				panel.add(reloadMixerButton, "wrap");
				reloadMixerButton.addActionListener(new ReloadList(this));
				
				mixerCombo=new JComboBox<String>();
				panel.add(mixerCombo, "span");
			}
			{
				startButton=new JButton("Start");
				leftPanel.add(startButton, "grow");
				startButton.addActionListener(new StartList(this));
			}
		}
		{
			outputText=new StandardOutputTextPane();
			JScrollPane scroll=new JScrollPane(outputText.textPane());
			scroll.setBorder(null);
			scroll.setPreferredSize(Dimensions.OUTPUT_TEXT);
			add(scroll, "wrap");
		}
	}

	public StandardOutputTextPane getOutputText() {
		return outputText;
	}

	public LinkedList<SoundFileLine> getSoundFileLine() {
		return soundFileLine;
	}

	public void addSoundFileLine()
	{
		Path prevDir=soundFileLine.getLast().getBrowseButton().getDirectory();
		SoundFileLine line=new SoundFileLine(this, prevDir, soundFilePanel);
		soundFileLine.add(line);
	}
	
	public void removeSoundFileLine(SoundFileLine line)
	{
		soundFileLine.remove(line);
		soundFilePanel.remove(line.getPathLabel());
		soundFilePanel.remove(line.getBrowseButton());
		soundFilePanel.remove(line.getRemoveButton());
		soundFilePanel.validate();
		soundFilePanel.repaint();
	}
	
	public void checkNewSoundFile(SoundFileLine target)
	{
		boolean dup=false;
		for(SoundFileLine line: soundFileLine) if(line!=target&&line.getBrowseButton().getSelected()!=null)
		{
			if(line.getBrowseButton().getSelected().toString().equals(target.getBrowseButton().getSelected().toString()))
				dup=true;
		}
		if(dup)
		{
			JOptionPane.showMessageDialog(this, "Duplicated files.", "Warning", JOptionPane.WARNING_MESSAGE);
		}
		
		boolean inconsistentFormat=false;
		for(SoundFileLine line: soundFileLine) if(line!=target&&line.getBrowseButton().getFormat()!=null)
		{
			if(!Utils.formatEquals(line.getBrowseButton().getFormat(), target.getBrowseButton().getFormat()))
				inconsistentFormat=true;
		}
		if(inconsistentFormat)
		{
			JOptionPane.showMessageDialog(this, "Inconsistent audio format.", "Warning", JOptionPane.WARNING_MESSAGE);
		}
		mainFrame.getMain().updateMixer(target.getBrowseButton().getFormat());
	}
	
	public void setMixer(List<Mixer> mixerInfo)
	{
		mixerCombo.removeAllItems();
		for(Mixer in: mixerInfo)
		{
//			mixerCombo.addItem(in.getMixerInfo().getName());
			String name=in.getMixerInfo().getName();
			if(System.getProperty("os.name").toLowerCase().contains("windows")&&Locale.getDefault().toString().equals("ja_JP"))
			{
				try
				{
					name=new String(name.getBytes("iso8859-1"), "Windows-31J");
				}
				catch (UnsupportedEncodingException e) {e.printStackTrace();}
			}
			mixerCombo.addItem(name);
		}
	}
	
	public ArrayList<Path> soundFile()
	{
		ArrayList<Path> soundFile=new ArrayList<Path>(soundFileLine.stream().map(s->s.getBrowseButton().getSelected()).filter(p->p!=null).collect(Collectors.toList()));
		return soundFile;
	}
	
	public void getConfigValue(Config config)
	{
		{
			Matcher m=INTEGER_PATTERN.matcher(silentText.getText());
			if(m.find())
			{
				long silentMs=Long.parseLong(silentText.getText());
				config.setSilentMs(silentMs);
			}
		}
		{
			Matcher m=INTEGER_PATTERN.matcher(numBlockText.getText());
			if(m.find())
			{
				int numBlock=Integer.parseInt(numBlockText.getText());
				config.setNumBlock(numBlock);
			}
		}
		if(logButton.getSelected()!=null) config.setLogDir(logButton.getDirectory());
		config.getAudioFile().clear();
		soundFileLine.forEach(line->config.getAudioFile().add(line.getBrowseButton().getSelected()));
		config.getAudioFile().removeIf(o->o==null);
	}
	
	public void setConfigValue(Config config)
	{
		silentText.setText(String.valueOf(config.getSilentMs()));
		numBlockText.setText(String.valueOf(config.getNumBlock()));
		logButton.setDirectory(config.getLogDir());
		config.getAudioFile().forEach(f->{
			soundFileLine.getLast().getPathLabel().setText(f.toString());
			soundFileLine.getLast().getBrowseButton().setSelected(f);
			soundFileLine.getLast().getBrowseButton().setDirectory(f.getParent());
			soundFileLine.getLast().setToolTipText();
			addSoundFileLine();
		});
		mainFrame.getMain().updateMixer(null);
	}
	
	public void stopped()
	{
		playing=false;
		startButton.setText("Start");
	}
	
	private static class StartList implements ActionListener
	{
		private MainPanel panel;
		
		private StartList(MainPanel panel) {
			this.panel = panel;
		}

		@Override
		public void actionPerformed(ActionEvent event)
		{
			if(panel.playing) stop();
			else play();
		}
		
		public void play()
		{
			LinkedList<String> errorText=new LinkedList<>();
			{
				Matcher m=INTEGER_PATTERN.matcher(panel.silentText.getText());
				if(!m.find()) errorText.add("Silent interval (ms) must be integer.");
			}
			{
				Matcher m=INTEGER_PATTERN.matcher(panel.numBlockText.getText());
				if(!m.find()) errorText.add("Repeat size must be integer.");
				else
				{
					int numBlock=Integer.parseInt(panel.numBlockText.getText());
					if(numBlock==0) errorText.add("Repeat size must be >0.");
				}
			}
			if(panel.logButton.getSelected()==null) errorText.add("Select log file.");
			if(panel.soundFileLine.size()==1) errorText.add("Select sound file.");
			AudioFormat format=null;
			boolean inconsistentFormat=false;
			for(Path path: panel.soundFile())
			{
				File file=path.toFile();
				if(!file.exists()) errorText.add(file+" does not exist.");
				else
				{
					AudioFormat f=null;
					try
					{
						f=AudioSystem.getAudioFileFormat(file).getFormat();
					}
					catch(UnsupportedAudioFileException e)
					{
						errorText.add("Unsupported audio file "+file+".");
					}
					catch(IOException e)
					{
						errorText.add("Could not read this file "+file+".");
					}
					if(format==null) format=f;
					else if(f!=null)
					{
						if(!Utils.formatEquals(format, f)) inconsistentFormat=true;
					}
				}
			}
			if(inconsistentFormat) errorText.add("Inconsistent audio format.");
			Mixer mixer=panel.mainFrame.getMain().mixer(panel.mixerCombo.getSelectedIndex());
			if(!Utils.isSupportedFormat(mixer, format))
			{
				errorText.add(mixer.getMixerInfo().getName()+" does not support selected sound format.");
			}
			
			if(errorText.size()>0)
			{
				StringBuffer buf=new StringBuffer("<html>");
				errorText.forEach(t->buf.append(t+"<br/>"));
				buf.append("</html>");
				Utils.errorMessage(panel, buf.toString());
				return;
			}
			
			if(panel.logButton.getSelected().toFile().exists())
			{
				int op=JOptionPane.showConfirmDialog(panel, "Log file exists. Overwrite?", "Warning", JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
				if(op!=JOptionPane.YES_OPTION) return;
			}
			
			Config config=new Config();
			panel.getConfigValue(config);
			StringBuffer buf=new StringBuffer("<html>Sound files will be played back with "+config.getSilentMs()+" ms silent intervals in the following order<br/><table>");
			ArrayList<Integer> order=Ordering.order(panel.mainFrame.getMain().random(), config.getNumBlock(), config.getAudioFile().size());
			for(int o=0; o<order.size(); ++o)
			{
				Path p=config.getAudioFile().get(order.get(o));
				buf.append("<tr><td>("+(o+1)+")</td><td>"+p+"</td></tr>");
			}
			buf.append("</table></html>");
			
			JScrollPane scroll=new JScrollPane(new JLabel(buf.toString()));
			scroll.setPreferredSize(new Dimension(600, 300));
			int answer=JOptionPane.showConfirmDialog(panel, scroll, "", JOptionPane.OK_CANCEL_OPTION);
			if(answer==JOptionPane.OK_OPTION)
			{
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						panel.playing=true;
						try {
							panel.mainFrame.getMain().play(config, order, mixer, panel.logButton.getSelected());
						} catch (IOException e) {
							e.printStackTrace();
						}
						panel.startButton.setText("Stop");
					}
				});
			}
		}
		
		public void stop()
		{
			try {
				panel.mainFrame.getMain().stop();
			} catch (InterruptedException | IOException e) {
				e.printStackTrace();
			}
		}
	}
	
	private static class ReloadList implements ActionListener
	{
		private MainPanel panel;
		
		private ReloadList(MainPanel panel) {
			this.panel = panel;
		}

		@Override
		public void actionPerformed(ActionEvent event)
		{
			panel.mainFrame.getMain().updateMixer(null);
		}
	}
}
